router id 192.0.2.21;

log "/var/log/bird.log" all;
log syslog all;
debug protocols all;

protocol static own_prefixes {
	route {{ data.AS2_pref1 }} reject;
	route {{ data.AS2_pref2 }} reject;
	route {{ data.AS2_pref3 }} reject;
	route {{ data.AS2_pref4 }} reject;
	route {{ data.AS2_pref5 }} reject;
}

protocol device { }
protocol bgp the_rs {
	local as 2;
	neighbor {{ data.rs_IPAddress }} as 999;
	next hop self;
	import all;
	export all;
	connect delay time 1;
	connect retry time 1;
}
